#define com_maroontress_libcgi_Request_IMPORT
#define com_maroontress_libcgi_db_Tree_IMPORT
#define com_maroontress_libcgi_db_Box_IMPORT
#define com_maroontress_libcgi_db_BoxKit_IMPORT
#define com_maroontress_libcgi_db_XML_IMPORT
#define com_maroontress_libcgi_util_Tub_IMPORT

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <libxml/parser.h>
#include <com/maroontress/libcgi/Request.h>
#include <com/maroontress/libcgi/db/Box.h>
#include <com/maroontress/libcgi/db/BoxKit.h>
#include <com/maroontress/libcgi/db/Tree.h>
#include <com/maroontress/libcgi/db/XML.h>
#include <com/maroontress/libcgi/util/Tub.h>
#include "BuildCharset.h"
#include "BuildDetail.h"
#include "BuildLog.h"
#include "Builder.h"
#include "Environment.h"
#include "Platform.h"
#include "PlatformList.h"
#include "Project.h"
#include "Serial.h"
#include "URL.h"
#include "script.h"

static void
ok(void)
{
    printf("\
Content-Type: text/xml\n\
\n\
<?xml version=\"1.0\"?>\n\
<status code=\"0\" message=\"Ok\"/>\n");
    exit(0);
}

static void
xmlFatalError(const char *format, ...)
{
    char *s, *xs;
    va_list ap;

    va_start(ap, format);
    vasprintf(&s, format, ap);
    va_end(ap);
    if (s != NULL) {
	xs = XML_createEscaped(s);
	free(s);
    }
    else {
	xs = NULL;
    }
    printf("\
Content-Type: text/xml\n\
\n\
<?xml version=\"1.0\"?>\n\
<status code=\"-1\" message=\"%s\"/>\n", xs);
    free(xs);
    exit(0);
}

#define DB_PROJECTS_DB "db/projects.db"
#define IMPORT_REQUEST(req, name)			\
    const char *name = Request_getValue(req, #name);
#define BIN_SH "/bin/sh"
#define BIN_CAT "/bin/cat"

static void
daemonFatalError(const char *format, ...)
{
    FILE *fp;
    va_list ap;

    va_start(ap, format);
    if ((fp = fopen("mad-angler.log", "a")) == NULL)
	_exit(0);
    vfprintf(fp, format, ap);
    fprintf(fp, "\n");
    fclose(fp);
    va_end(ap);
    _exit(0);
}

static int
hasNextBuildWithProjects(struct Builder *b, int result, struct Tree *projects)
{
    struct BoxKit *project;
    struct Serial *serial = Builder_getSerial(b);
    char *projectID = Builder_getProjectID(b);
    int currentSerial;

    if ((project = Tree_getBoxKit(projects, projectID)) == NULL) {
	daemonFatalError("Tree_getBoxKit() failed.");
    }
    if (BoxKit_getInt(project, "serial", &currentSerial) < 0) {
	daemonFatalError("BoxKit_getInt() failed.");
    }
    if (currentSerial > Serial_getInt(serial)) {
	if (Builder_setSerial(b, currentSerial) < 0) {
	    daemonFatalError("BoxKit_getInt() failed.");
	}
	BoxKit_purge(project);
	return 1;
    }
    if (result == Builder_FAIL && Builder_findPrevious(b)) {
	BoxKit_purge(project);
	return 1;
    }
    if (BoxKit_setInt(project, "load", 0) < 0
	|| BoxKit_close(project) < 0) {
	daemonFatalError("can't set load value to 0.");
    }
    return 0;
}

static int
hasNextBuild(struct Builder *b, int result)
{
    struct Tree *projects;
    int s;
    
    if ((projects = Tree_create(DB_PROJECTS_DB,
				Tree_MODE_EXCLUSIVE)) == NULL) {
	daemonFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    }
    s = hasNextBuildWithProjects(b, result, projects);
    Tree_free(projects);
    return s;
}

static int
sendReportMail(struct Builder *b, int status, struct Environment *e)
{
    static const char *const COMMAND[] = {BIN_SH, "post-all-build.sh", NULL};
    char *projectID = Builder_getProjectID(b);
    char *revision = Builder_getLastRevision(b);
    char *urlEncodedProjectId;
    int serial = Serial_getInt(Builder_getSerial(b));
    int pid;
    char buf[16];

    if ((pid = fork()) < 0)
	return -1;
    if (pid != 0)
	return 0;
    if (Builder_getMailEnabled(b)
	&& (Environment_setRaw(e, "HTTP_HOST", getenv("HTTP_HOST")) < 0
	    || Environment_setRaw(e, "SERVER_PORT", getenv("SERVER_PORT")) < 0
	    || Environment_setRaw(e, "SCRIPT_NAME", getenv("SCRIPT_NAME")) < 0
	    || Environment_setRaw(e, "MAIL_ENABLED", "YES") < 0)) {
	daemonFatalError("Environment_setRaw() failed.");
    }
    snprintf(buf, sizeof(buf), "%d", serial);
    urlEncodedProjectId = URL_encodeComponent(projectID);
    if (Environment_setRaw(e, "projectId", projectID) < 0
	|| Environment_setRaw(e, "urlEncodedProjectId",
			      urlEncodedProjectId) < 0
	|| Environment_setRaw(e, "result",
			      (status == Builder_SUCCESS) ? "OK" : "NG") < 0
	|| Environment_setRaw(e, "revision", revision) < 0
	|| Environment_setRaw(e, "serial", buf) < 0
	|| Environment_execute(e, COMMAND) < 0) {
	daemonFatalError("Environment_execute() failed.");
    }
    free(urlEncodedProjectId);
    _exit(0);
}

static void
build(struct Builder *b, struct Environment *e)
{
    char *projectID = Builder_getProjectID(b);
    int status;

    for (;;) {
	if ((status = Builder_run(b, e)) < 0) {
	    daemonFatalError("Builder_run() failed: %s", projectID);
	}
	if (sendReportMail(b, status, e) < 0) {
	    daemonFatalError("sendReportMail() failed: %s", projectID);
	}
	if (!hasNextBuild(b, status)) {
	    break;
	}
    }
    _exit(0);
}

static void
serveGetMethod(struct Request *req, struct Environment *e)
{
    IMPORT_REQUEST(req, projectId);
    IMPORT_REQUEST(req, revision);
    IMPORT_REQUEST(req, author);
    struct Tree *projects;
    struct Box *box;
    struct BoxKit *kit;
    int pid, platforms;
    struct Builder *b;
    
    if (projectId == NULL || revision == NULL || author == NULL)
	xmlFatalError("bad request.");
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_EXCLUSIVE)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((box = Tree_getBox(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    if (Environment_setRawByBox(e, "SMTP_SERVER", box, "smtpServer") < 0
	|| Environment_setRawByBox(e, "MAIL_FROM", box, "sender") < 0
	|| Environment_setRawByBox(e, "RCPT_TO", box, "receipient") < 0) {
	xmlFatalError("can't set mail properties.");
    }
    if ((kit = Tree_getBoxKit(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    if (BoxKit_getInt(kit, "platforms", &platforms) < 0
	|| platforms == 0) {
	xmlFatalError("no platforms found: ", projectId);
    }
    if ((b = Builder_create(kit, revision, author)) == NULL) {
	xmlFatalError("Builder_create() failed.");
    }
    if (BoxKit_close(kit) < 0) {
	/* build.dbΥХåɬ */
	xmlFatalError("failed to update: %s: %s", DB_PROJECTS_DB, projectId);
    }
    Tree_free(projects);

    if (Builder_getLoad(b) == 0) {
        if ((pid = fork()) < 0) {
	    xmlFatalError("fork() failed.");
        }
        if (pid == 0) {
            close(0);
            close(1);
            close(2);
            build(b, e);
        }
    }
    Builder_free(b);
    ok();
}

static int
writeTrimmedKey(struct Tub *tub, struct Box *box)
{
    const char *key = Box_getKey(box);
    char buf[16];

    snprintf(buf, sizeof(buf), "%d", atoi(key));
    return Tub_writeCString(tub, buf);
}

static int
writeXMLPreviousElement(struct Tub *tub, struct Tree *tree,
			const char *startKey)
{
    struct Box *box;
    unsigned int k;

    for (box = Tree_scanBox(tree, startKey), k = 0;
         box != NULL && k < 10;
         box = Tree_scanNextBox(tree), ++k) {
        continue;
    }
    if (Tree_hasError(tree))
        return -1;
    if (box == NULL)
        return 0;

    if (Tub_writeCString(tub, "<prev serial=\"") < 0
	|| writeTrimmedKey(tub, box) < 0
	|| Tub_writeCString(tub, "\"/>\n") < 0) {
	return -1;
    }
    return 0;
}

static int
writeXMLBuildElement(struct Tub *tub, struct Box *box)
{
    char *revision, *author;
    time_t t;
    int stamp, result;
    char date[32], status[8];

    if (Box_getString(box, "revision", &revision) < 0
	|| Box_getString(box, "author", &author) < 0
	|| Box_getInt(box, "stamp", &stamp) < 0
	|| Box_getInt(box, "result", &result) < 0) {
	return -1;
    }

    t = (time_t)stamp;
    strftime(date, sizeof(date), "%Y-%m-%d %T", localtime(&t));
    snprintf(status, sizeof(status), "%d", result);

    if (Tub_writeCString(tub, "<build serial=\"") < 0
	|| writeTrimmedKey(tub, box) < 0
	|| Tub_writeCString(tub, "\" revision=\"") < 0
	|| XML_writeEscapedToTub(tub, revision) < 0
	|| Tub_writeCString(tub, "\" author=\"") < 0
	|| XML_writeEscapedToTub(tub, author) < 0
	|| Tub_writeCString(tub, "\" date=\"") < 0
	|| XML_writeEscapedToTub(tub, date) < 0
	|| Tub_writeCString(tub, "\" status=\"") < 0
	|| Tub_writeCString(tub, status) < 0
	|| Tub_writeCString(tub, "\"/>\n") < 0) {
	return -1;
    }
    return 0;
}

static int
writeBuildList(struct Tub *tub, struct Tree *tree, struct Serial *start)
{
    struct Box *box;
    unsigned int k;

    if (Tub_writeCString(tub, "<status code=\"0\" message=\"OK\">\n") < 0)
	return -1;
    if (start == NULL) {
	box = Tree_scanLastBox(tree);
    }
    else {
	char *startKey = Serial_getKey(start);
	if (writeXMLPreviousElement(tub, tree, startKey) < 0)
	    return -1;
	box = Tree_scanBox(tree, startKey);
    }
    for (k = 0; box != NULL && k < 10; ++k, box = Tree_scanPreviousBox(tree)) {
	if (writeXMLBuildElement(tub, box) < 0) {
	    return -1;
	}
    }
    if (Tree_hasError(tree)
	|| (box != NULL
	    && (Tub_writeCString(tub, "<next serial=\"") < 0
		|| writeTrimmedKey(tub, box) < 0
		|| Tub_writeCString(tub, "\"/>\n") < 0))
	|| Tub_writeCString(tub, "</status>") < 0
	|| Tub_writeByte(tub, 0) < 0) {
	return -1;
    }
    return 0;
}

static char *
createBuildListWithSerial(struct Tree *tree, struct Serial *start)
{
    struct Tub *tub;
    char *bl;

    if ((tub = Tub_create()) == NULL)
	return NULL;
    bl = ((writeBuildList(tub, tree, start) < 0)
	  ? NULL : Tub_createByteArray(tub));
    Tub_free(tub);
    return bl;
}

static char *
createBuildList(struct Tree *tree, const char *startKey)
{
    struct Serial *serial;
    char *xml;

    if (startKey == NULL) {
	return createBuildListWithSerial(tree, NULL);
    }
    if ((serial = Serial_create(atoi(startKey))) == NULL)
	return NULL;
    xml = createBuildListWithSerial(tree, serial);
    Serial_free(serial);
    return xml;
}

static void
Method_listBuilds(struct Request *req UNUSED, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    IMPORT_REQUEST(req, startKey);
    struct Project *pj;
    struct Tree *projects, *builds;
    struct Box *project;
    int serial;
    char *buildList;

    if (projectId == NULL)
	xmlFatalError("bad request.");
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_SHARED)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((project = Tree_getBox(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    if (Box_getInt(project, "serial", &serial) < 0)
        xmlFatalError("projectId not found: %s", projectId);
    Tree_free(projects);

    if (serial == 0) {
	printf("Content-type: text/xml\n\n"
	       "<status code=\"0\" message=\"OK\"/>\n");
	return;
    }
    if ((pj = Project_create(projectId)) == NULL)
	xmlFatalError("Project_create() failed: %s", projectId);
    if ((builds = Project_createTree(pj, Tree_MODE_SHARED)) == NULL) {
	xmlFatalError("Tree_create() failed: %s", projectId);
    }
    if ((buildList = createBuildList(builds, startKey)) == NULL)
	xmlFatalError("createProjcetLists() failed: %s", DB_PROJECTS_DB);
    Tree_free(builds);
    Project_free(pj);
    printf("Content-type: text/xml\n\n%s\n", buildList);
    free(buildList);
}

static int
writeProjectList(struct Tub *tub, struct Tree *tree)
{
    struct Box *box;

    if (Tub_writeCString(tub, "<projectList>\n") < 0)
	return -1;
    for (box = Tree_scanFirstBox(tree);
	 box != NULL;
	 box = Tree_scanNextBox(tree)) {
	if (Box_writeXMLElementWithKeyToTub(box, tub, "project", "name",
					    "description", NULL) < 0
	    || Tub_writeByte(tub, '\n') < 0) {
	    return -1;
	}
    }
    if (Tree_hasError(tree)
	|| Tub_writeCString(tub, "</projectList>") < 0
	|| Tub_writeByte(tub, 0) < 0) {
	return -1;
    }
    return 0;
}

static char *
createProjcetList(struct Tree *tree)
{
    struct Tub *tub;
    char *pl;

    if ((tub = Tub_create()) == NULL)
	return NULL;
    pl = (writeProjectList(tub, tree) < 0) ? NULL : Tub_createByteArray(tub);
    Tub_free(tub);
    return pl;
}

static void
Method_listProjects(struct Request *req UNUSED, struct Environment *e UNUSED)
{
    struct Tree *projects;
    char *projs;

    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_SHARED)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((projs = createProjcetList(projects)) == NULL)
	xmlFatalError("createProjcetList() failed: %s", DB_PROJECTS_DB);
    Tree_free(projects);
    printf("Content-type: text/xml\n\n%s\n", projs);
    free(projs);
}

static void
Method_getBuildDetail(struct Request *req, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    IMPORT_REQUEST(req, serial);
    struct BuildDetail *detail;

    if (projectId == NULL || serial == NULL)
	xmlFatalError("bad request.");
    if ((detail = BuildDetail_create(projectId, serial)) == NULL)
	xmlFatalError("BuildDetail_create() failed.");
    printf("Content-type: text/xml\n\n%s\n", BuildDetail_getResponse(detail));
    BuildDetail_free(detail);
}

static int
writePropertyResponse(struct Box *box, struct Tub *tub)
{
    int k, platforms;
    char *id, *host, *command, *charset;

    if (Box_getInt(box, "platforms", &platforms) < 0
	|| Tub_writeCString(tub, "<status code=\"0\" message=\"OK\">\n") < 0
	|| Box_writeXMLElementToTub(box, tub, "property",
				    "description",
				    "mailEnabled",
				    "smtpServer",
				    "sender",
				    "receipient", NULL) < 0
	|| Tub_writeCString(tub, "\n") < 0) {
	return -1;
    }
    for (k = 0; k < platforms; ++k) {
	if (Platform_getStringWithBox(box, k, "id", &id) < 0
	    || Platform_getStringWithBox(box, k, "host", &host) < 0
	    || Platform_getStringWithBox(box, k, "command", &command) < 0
	    || Platform_getStringWithBox(box, k, "charset", &charset) < 0
	    || Tub_writeCString(tub, "<platform id=\"") < 0
	    || XML_writeEscapedToTub(tub, id) < 0
	    || Tub_writeCString(tub, "\" host=\"") < 0
	    || XML_writeEscapedToTub(tub, host) < 0
	    || Tub_writeCString(tub, "\" command=\"") < 0
	    || XML_writeEscapedToTub(tub, command) < 0
	    || Tub_writeCString(tub, "\" charset=\"") < 0
	    || XML_writeEscapedToTub(tub, charset) < 0
	    || Tub_writeCString(tub, "\"/>\n") < 0) {
	    return -1;
	}
    }
    if (Tub_writeCString(tub, "</status>\n") < 0
	|| Tub_writeByte(tub, 0) < 0) {
	return -1;
    }
    return 0;
}

static char *
createPropertyResponse(struct Box *box)
{
    struct Tub *tub;
    char *res;

    if ((tub = Tub_create()) == NULL)
	return NULL;
    res = ((writePropertyResponse(box, tub) < 0)
	   ? NULL : Tub_createByteArray(tub));
    Tub_free(tub);
    return res;
}

static void
Method_getProperty(struct Request *req, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    struct Tree *projects;
    struct Box *project;
    char *res;

    if (projectId == NULL)
	xmlFatalError("bad request.");
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_SHARED)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((project = Tree_getBox(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    if ((res = createPropertyResponse(project)) == NULL)
	xmlFatalError("too short memory.");
    Tree_free(projects);
    printf("Content-type: text/xml\n\n%s\n", res);
    free(res);
}

static int
ProjectConfig_copyAttribute(struct BoxKit *project, xmlNodePtr node,
			    const char *name)
{
    char *value;
    int s;

    if ((value = xmlGetProp(node, name)) == NULL)
	return -1;
    s = BoxKit_setString(project, name, value);
    xmlFree(value);
    return s;
}

static int
ProjectConfig_copyIntAttribute(struct BoxKit *project, xmlNodePtr node,
			       const char *name)
{
    char *value;
    int s;

    if ((value = xmlGetProp(node, name)) == NULL)
	return -1;
    s = BoxKit_setInt(project, name, atoi(value));
    xmlFree(value);
    return s;
}

static int
ProjectConfig_copyPlatformAttribute(struct BoxKit *project, int k,
				    xmlNodePtr node, const char *name)
{
    char *value;
    int s;

    if ((value = xmlGetProp(node, name)) == NULL)
	return -1;
    s = Platform_setStringWithBoxKit(project, k, name, value);
    xmlFree(value);
    return s;
}

static int
setProjectConfigProperty(struct BoxKit *project, xmlNodePtr node)
{
    if (ProjectConfig_copyAttribute(project, node, "description") < 0
	|| ProjectConfig_copyIntAttribute(project, node, "mailEnabled") < 0
	|| ProjectConfig_copyAttribute(project, node, "smtpServer") < 0
	|| ProjectConfig_copyAttribute(project, node, "sender") < 0
	|| ProjectConfig_copyAttribute(project, node, "receipient") < 0) {
	return -1;
    }
    return 0;
}

static int
setProjectConfigPlatform(struct BoxKit *kit, xmlNodePtr node, int k)
{
    if (ProjectConfig_copyPlatformAttribute(kit, k, node, "id") < 0
	|| ProjectConfig_copyPlatformAttribute(kit, k, node, "host") < 0
	|| ProjectConfig_copyPlatformAttribute(kit, k, node, "command") < 0
	|| ProjectConfig_copyPlatformAttribute(kit, k, node, "charset") < 0) {
	return -1;
    }
    return 0;
}

static int
setProjectConfigDocument(struct BoxKit *project, xmlDocPtr doc)
{
    xmlNodePtr root, c;
    int k;

    if ((root = xmlDocGetRootElement(doc)) == NULL
	|| strcmp(root->name, "projectConfig") != 0) {
	return -1;
    }

    for (c = root->xmlChildrenNode; c != NULL; c = c->next) {
        if (c->type != XML_ELEMENT_NODE
	    || strcmp(c->name, "property") != 0) {
            continue;
	}
	if (setProjectConfigProperty(project, c) < 0) {
	    return -1;
	}
    }

    k = 0;
    for (c = root->xmlChildrenNode; c != NULL; c = c->next) {
        if (c->type != XML_ELEMENT_NODE
	    || strcmp(c->name, "platform") != 0) {
            continue;
	}
	if (setProjectConfigPlatform(project, c, k) < 0) {
	    return -1;
	}
	++k;
    }
    return BoxKit_setInt(project, "platforms", k);
}

static int
setProjectConfig(struct BoxKit *project, const char *projectConfig)
{
    xmlDocPtr doc;
    int s;

    if ((doc = xmlParseMemory(projectConfig, strlen(projectConfig))) == NULL)
	return -1;
    s = setProjectConfigDocument(project, doc);
    xmlFreeDoc(doc);
    return s;
}

static void
Method_setProperty(struct Request *req, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    IMPORT_REQUEST(req, projectConfig);
    struct Tree *projects;
    struct BoxKit *project;

    if (projectId == NULL)
	xmlFatalError("bad request.");
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_EXCLUSIVE)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((project = Tree_getBoxKit(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    /*
      Υץѥƥ򤹤٤ƾõʤȻȤʤեɤĤ
      BoxKit_clear(project);
      ߤΤɬס
    */
    if (setProjectConfig(project, projectConfig) < 0)
	xmlFatalError("setProjectConfig() failed.");
    if (BoxKit_close(project) < 0)
	xmlFatalError("BoxKit_close() failed: %s", projectId);
    Tree_free(projects);
    ok();
}

static void
Method_deleteProject(struct Request *req, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    struct Project *pj;
    struct Tree *projects;

    if (projectId == NULL)
	xmlFatalError("bad request.");
    if ((pj = Project_create(projectId)) == NULL)
        xmlFatalError("Project_create() failed: %s", projectId);
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_EXCLUSIVE)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if (Tree_delete(projects, projectId) != 0)
        xmlFatalError("projectId not exists: %s", projectId);
    if (Project_removeDir(pj) < 0)
        xmlFatalError("Project_removeDir() failed: %s", projectId);
    Tree_free(projects);
    Project_free(pj);
    ok();
}

static void
Method_createProject(struct Request *req, struct Environment *e UNUSED)
{
    IMPORT_REQUEST(req, projectId);
    struct Project *pj;
    struct Tree *projects;
    struct BoxKit *project;
    int serial;

    if (projectId == NULL)
	xmlFatalError("bad request.");
    if (strlen(projectId) == 0)
	xmlFatalError("Project ID is empty.");
    if ((projects = Tree_create(DB_PROJECTS_DB, Tree_MODE_EXCLUSIVE)) == NULL)
	xmlFatalError("Tree_create() failed: %s", DB_PROJECTS_DB);
    if ((project = Tree_getBoxKit(projects, projectId)) == NULL)
	xmlFatalError("bad projectId parameter.");
    if (BoxKit_getInt(project, "serial", &serial) == 0)
        xmlFatalError("projectId already exists: %s", projectId);
    if (BoxKit_setInt(project, "serial", 0) < 0 
        || BoxKit_setInt(project, "load", 0) < 0
        || BoxKit_setInt(project, "platforms", 0) < 0
        || BoxKit_setString(project, "description", "") < 0
        || BoxKit_setInt(project, "mailEnabled", 0) < 0
        || BoxKit_setString(project, "smtpServer", "localhost") < 0
        || BoxKit_setString(project, "sender", "") < 0
        || BoxKit_setString(project, "receipient", "") < 0) {
	BoxKit_purge(project);
	xmlFatalError("can't create project: %s", projectId);
    }
    if ((pj = Project_create(projectId)) == NULL)
        xmlFatalError("Project_create() failed: %s", projectId);
    if (Project_makeDir(pj) < 0)
	xmlFatalError("Project_makeDir() failed: %s", projectId);
    if (BoxKit_close(project) < 0) {
	Project_removeDir(pj);
	xmlFatalError("BoxKit_close() failed: %s", projectId);
    }
    Project_free(pj);
    Tree_free(projects);
    ok();
}

static void
Method_getPublicKey(struct Request *req UNUSED, struct Environment *e)
{
    static const char *const COMMAND[] = {BIN_CAT, "id_rsa.pub.txt", NULL};

    if ((e = Environment_create()) == NULL)
	xmlFatalError("Environment_create() failed.");
    if (Environment_execute(e, COMMAND) < 0)
	xmlFatalError("Environment_execute() failed: %s", COMMAND[0]);
}

static void
Method_getBuildLog(struct Request *req, struct Environment *e)
{
    static const char *const COMMAND[] = {BIN_SH, "log.sh", NULL};
    IMPORT_REQUEST(req, projectId);
    IMPORT_REQUEST(req, serial);
    IMPORT_REQUEST(req, platformCode);
    IMPORT_REQUEST(req, tail);
    struct BuildCharset *cs;
    struct BuildLog *bl;
    struct Serial *sr;
    char path[MAXPATHLEN];

    if (projectId == NULL || serial == NULL || platformCode == NULL)
	xmlFatalError("bad request.");
    if (Environment_setRaw(e, "ICONV", ICONV) < 0)
	xmlFatalError("Environment_setRaw() failed: ICONV: %s", ICONV);
    if (Environment_copyXML(e, projectId) < 0
	|| Environment_copyXML(e, serial) < 0
	|| Environment_copyXML(e, platformCode) < 0
	|| Environment_copyXML(e, tail) < 0) {
	xmlFatalError("Environment_copyXML() failed.");
    }

    if ((cs = BuildCharset_create(projectId, serial, platformCode)) == NULL)
	xmlFatalError("BuildCharset_create() failed.");
    if (Environment_setXML(e, "charset", BuildCharset_getCharset(cs)) < 0)
	xmlFatalError("Environment_setXML() failed.");
    BuildCharset_free(cs);

    if ((sr = Serial_create(atoi(serial))) == NULL)
	xmlFatalError("Serial_create() failed: %s", serial);
    if ((bl = BuildLog_create(projectId)) == NULL)
	xmlFatalError("BuildLog_create() failed: %s", projectId);
    BuildLog_getBasePath(bl, sr, path, sizeof(path));
    if (Environment_setRaw(e, "logDir", path) < 0) {
	xmlFatalError("Environment_setRaw() failed: %s: %s",
		      "logDir", path);
    }
    BuildLog_free(bl);
    Serial_free(sr);

    if (Environment_execute(e, COMMAND) < 0)
	xmlFatalError("Environment_execute() failed: %s", COMMAND[0]);
}

#define METHOD_DEFINE(x) {#x, Method_##x}
static const struct {
    const char *name;
    void (*run)(struct Request *, struct Environment *);
} methodTable[] = {
    METHOD_DEFINE(listProjects),
    METHOD_DEFINE(createProject),
    METHOD_DEFINE(deleteProject),
    METHOD_DEFINE(getProperty),
    METHOD_DEFINE(setProperty),
    METHOD_DEFINE(listBuilds),
    METHOD_DEFINE(getBuildDetail),
    METHOD_DEFINE(getBuildLog),
    METHOD_DEFINE(getPublicKey),
    /* methodTable */};

int
main(void)
{
    struct Request *req;
    struct Environment *e;
    const char *method;
    unsigned int k, n;

    if (chdir(PROJECTDIR) < 0) {
	xmlFatalError("bad configuration.");
    }
    if ((req = Request_create()) == NULL) {
	xmlFatalError("bad request.");
    }
    if ((e = Environment_create()) == NULL) {
	xmlFatalError("Environment_create() failed.");
    }
    if ((method = Request_getValue(req, "method")) == NULL) {
	/* GET᥽åɤˤꥯȤβǽ */
	serveGetMethod(req, e);
	exit(0);
    }
    n = sizeof(methodTable)/sizeof(methodTable[0]);
    for (k = 0; k < n; ++k) {
	if (strcmp(method, methodTable[k].name) == 0) {
	    methodTable[k].run(req, e);
	    break;
	}
    }
    if (k == n) {
	xmlFatalError("invalid method parameter");
    }
    Environment_free(e);
    exit(0);
    return 0;
}
